@@ -6,11 +6,12 @@ const Tp = imports.gi.TelepathyGLib;
 
 const AccountsMonitor = imports.accountsMonitor;
 const AppNotifications = imports.appNotifications;
-const ChatroomManager = imports.chatroomManager;
 const Connections = imports.connections;
 const Lang = imports.lang;
 const MainWindow = imports.mainWindow;
 const PasteManager = imports.pasteManager;
+const RoomManager = imports.roomManager;
+const TelepathyClient = imports.telepathyClient;
 const Utils = imports.utils;
 const NetworksManager = imports.networksManager;
 
@@ -18,19 +19,11 @@ const MAX_RETRIES = 3;
 
 const IRC_SCHEMA_REGEX = /^(irc?:\/\/)([\da-z\.-]+):?(\d+)?\/(?:%23)?([\w\.\+-]+)/i;
 
-const ConnectionError = {
-    CANCELLED: Tp.error_get_dbus_name(Tp.Error.CANCELLED),
-    ALREADY_CONNECTED: Tp.error_get_dbus_name(Tp.Error.ALREADY_CONNECTED),
-    DISCONNECTED: Tp.error_get_dbus_name(Tp.Error.DISCONNECTED),
-    NETWORK_ERROR: Tp.error_get_dbus_name(Tp.Error.NETWORK_ERROR),
-    NOT_AVAILABLE: Tp.error_get_dbus_name(Tp.Error.NOT_AVAILABLE),
-    SERVICE_BUSY: Tp.error_get_dbus_name(Tp.Error.SERVICE_BUSY)
-};
-
 const Application = new Lang.Class({
     Name: 'Application',
     Extends: Gtk.Application,
-    Signals: { 'prepare-shutdown': {} },
+    Signals: { 'prepare-shutdown': {},
+               'room-focus-changed': {} },
 
     _init: function() {
         this.parent({ application_id: 'org.gnome.Polari',
@@ -38,28 +31,18 @@ const Application = new Lang.Class({
 
         GLib.set_application_name('Polari');
         this._window = null;
-        this._pendingRequests = {};
+        this._retryData = new Map();
+    },
+
+    isRoomFocused: function(room) {
+        return this.active_window &&
+               this.active_window.is_active &&
+               this.active_window.active_room == room;
     },
 
     vfunc_startup: function() {
         this.parent();
 
-        this._chatroomManager = ChatroomManager.getDefault();
-        this._accountsMonitor = AccountsMonitor.getDefault();
-        this._networkMonitor = Gio.NetworkMonitor.get_default();
-        this._networksManager = NetworksManager.getDefault();
-
-        this._accountsMonitor.connect('account-removed', Lang.bind(this,
-            function(am, account) {
-                this._removeSavedChannelsForAccount(account);
-            }));
-
-        this._settings = new Gio.Settings({ schema_id: 'org.gnome.Polari' });
-
-        this.pasteManager = new PasteManager.PasteManager();
-        this.notificationQueue = new AppNotifications.NotificationQueue();
-        this.commandOutputQueue = new AppNotifications.CommandOutputQueue();
-
         let actionEntries = [
           { name: 'show-join-dialog',
             activate: Lang.bind(this, this._onShowJoinDialog),
@@ -71,15 +54,18 @@ const Application = new Lang.Class({
             activate: Lang.bind(this, this._onMessageUser),
             parameter_type: GLib.VariantType.new('(sssu)') },
           { name: 'leave-room',
-            activate: Lang.bind(this, this._onLeaveRoom),
             parameter_type: GLib.VariantType.new('(ss)') },
           { name: 'leave-current-room',
             activate: Lang.bind(this, this._onLeaveCurrentRoom),
-            create_hook: Lang.bind(this, this._leaveRoomCreateHook),
+            create_hook: (a) => { a.enabled = false; },
             accels: ['<Primary>w'] },
           { name: 'authenticate-account',
             parameter_type: GLib.VariantType.new('(os)') },
+          { name: 'connect-account',
+            activate: Lang.bind(this, this._onConnectAccount),
+            parameter_type: GLib.VariantType.new('o') },
           { name: 'reconnect-account',
+            activate: Lang.bind(this, this._onConnectAccount),
             parameter_type: GLib.VariantType.new('o') },
           { name: 'user-list',
             activate: Lang.bind(this, this._onToggleAction),
@@ -92,6 +78,10 @@ const Application = new Lang.Class({
           { name: 'edit-connection',
             activate: Lang.bind(this, this._onEditConnection),
             parameter_type: GLib.VariantType.new('o') },
+          { name: 'save-identify-password',
+            parameter_type: GLib.VariantType.new('o') },
+          { name: 'discard-identify-password',
+            parameter_type: GLib.VariantType.new('o') },
           { name: 'help',
             activate: Lang.bind(this, this._onShowHelp),
             accels: ['F1'] },
@@ -138,47 +128,73 @@ const Application = new Lang.Class({
 
         for (let i = 1; i < 10; i++)
             this.set_accels_for_action('app.nth-room(%d)'.format(i), ['<Alt>' + i]);
+
+        this._telepathyClient = null;
+
+        this._roomManager = RoomManager.getDefault();
+        this._accountsMonitor = AccountsMonitor.getDefault();
+        this._networksManager = NetworksManager.getDefault();
+
+        this._accountsMonitor.connect('account-status-changed',
+                                      Lang.bind(this, this._onAccountStatusChanged));
+
+        this.pasteManager = new PasteManager.PasteManager();
+        this.notificationQueue = new AppNotifications.NotificationQueue();
+        this.commandOutputQueue = new AppNotifications.CommandOutputQueue();
     },
 
     vfunc_activate: function() {
+        if (!this._telepathyClient) {
+            let params = {
+                name: 'Polari',
+                account_manager: this._accountsMonitor.accountManager,
+                uniquify_name: false
+            };
+            this._telepathyClient = new TelepathyClient.TelepathyClient(params);
+        }
+
         if (!this._window) {
             this._window = new MainWindow.MainWindow({ application: this });
-            this._window.connect('destroy', Lang.bind(this,
-                function() {
-                    for (let id in this._pendingRequests)
-                        this._pendingRequests[id].cancellable.cancel();
-                    this.emit('prepare-shutdown');
-            }));
+            this._window.connect('destroy',
+                                 () => { this.emit('prepare-shutdown'); });
+            this._window.connect('notify::active-room',
+                                 () => { this.emit('room-focus-changed'); });
+            this._window.connect('notify::is-active',
+                                 () => { this.emit('room-focus-changed'); });
             this._window.show_all();
-
-            this._chatroomManager.lateInit();
         }
         this._window.present();
     },
 
+    vfunc_window_added: function(window) {
+        this.parent(window);
+
+        let action = this.lookup_action('leave-current-room');
+        window.connect('notify::active-room', () => {
+            action.enabled = window.active_room != null;
+        });
+        action.enabled = window.active_room != null;
+
+        window.connect('active-room-state-changed',
+                       Lang.bind(this, this._updateUserListAction));
+        this._updateUserListAction();
+    },
+
     vfunc_open: function(files) {
         this.activate();
 
         let time = Utils.getTpEventTime();
         let uris = files.map(function(f) { return f.get_uri(); });
 
-        let quark = Tp.AccountManager.get_feature_quark_core();
-        if (this._accountsMonitor.accountManager.is_prepared(quark))
+        this._accountsMonitor.prepare(() => {
             this._openURIs(uris, time);
-        else
-            this._accountsMonitor.connect('account-manager-prepared', Lang.bind(this,
-                function(mon) {
-                    this._openURIs(uris, time);
-                }));
+        });
     },
 
     _openURIs: function(uris, time) {
         let map = {};
 
-        this._accountsMonitor.dupAccounts().forEach(function(a) {
-            if (!a.enabled)
-                return;
-
+        this._accountsMonitor.enabledAccounts.forEach(a => {
             let params = a.dup_parameters_vardict().deep_unpack();
             map[a.get_object_path()] = {
                 server: params.server.deep_unpack(),
@@ -192,10 +208,6 @@ const Application = new Lang.Class({
             if (!success)
                 return;
 
-            // When launching Polari via a URL, there is a potential race
-            // condition between networksManager and AccountsManager here. If
-            // AccountsManager finishes first, networksManager's list of
-            // servers would be empty and hence no matches would be found.
             let matchedId = this._networksManager.findByServer(server);
             let matches = Object.keys(map).filter(function(a) {
                 return GLib.ascii_strcasecmp(map[a].server, server) == 0 ||
@@ -267,293 +279,157 @@ const Application = new Lang.Class({
             }));
     },
 
-    _updateAccountAction: function(action) {
-        action.enabled = this._accountsMonitor.dupAccounts().filter(
-            function(a) {
-                return a.enabled;
-            }).length > 0;
-    },
-
-    _leaveRoomCreateHook: function(action) {
-        this._chatroomManager.connect('active-changed', Lang.bind(this,
-            function() {
-                action.enabled = this._chatroomManager.getActiveRoom() != null;
-            }));
-        action.enabled = this._chatroomManager.getActiveRoom() != null;
-    },
-
-    _updateUserListAction: function(action) {
-        let room = this._chatroomManager.getActiveRoom();
+    _updateUserListAction: function() {
+        let room = this.active_window.active_room;
+        let action = this.lookup_action('user-list');
         action.enabled = room && room.type == Tp.HandleType.ROOM && room.channel;
     },
 
     _userListCreateHook: function(action) {
-        this._chatroomManager.connect('active-state-changed', Lang.bind(this,
-            function() {
-                this._updateUserListAction(action);
-            }));
         action.connect('notify::enabled', function() {
             if (!action.enabled)
                 action.change_state(GLib.Variant.new('b', false));
         });
-        this._updateUserListAction(action);
+        action.enabled = false;
     },
 
     _onShowJoinDialog: function() {
         this._window.showJoinRoomDialog();
     },
 
-    _savedChannelIndex: function(savedChannels, account, channel) {
-        let accountPath = account.get_object_path();
-        let matchChannel = channel.toLowerCase();
-        for (let i = 0; i < savedChannels.length; i++)
-            if (savedChannels[i].account.deep_unpack() == accountPath &&
-                savedChannels[i].channel.deep_unpack().toLowerCase() == matchChannel)
-                return i;
-        return -1;
-    },
+    _maybePresent: function(time) {
+        let [present, ] = Tp.user_action_time_should_present(time);
 
-    _addSavedChannel: function(account, channel) {
-        let savedChannels = this._settings.get_value('saved-channel-list').deep_unpack();
-        if (this._savedChannelIndex(savedChannels, account, channel) != -1)
-            return;
-        savedChannels.push({
-            account: GLib.Variant.new('s', account.get_object_path()),
-            channel: GLib.Variant.new('s', channel)
-        });
-        this._settings.set_value('saved-channel-list',
-                                 GLib.Variant.new('aa{sv}', savedChannels));
-    },
-
-    _removeSavedChannel: function(account, channel) {
-        let savedChannels = this._settings.get_value('saved-channel-list').deep_unpack();
-        let pos = this._savedChannelIndex(savedChannels, account, channel);
-        if (pos < 0)
-            return;
-        savedChannels.splice(pos, 1);
-        this._settings.set_value('saved-channel-list',
-                                 GLib.Variant.new('aa{sv}', savedChannels));
-    },
-
-    _removeSavedChannelsForAccount: function(account) {
-        let savedChannels = this._settings.get_value('saved-channel-list').deep_unpack();
-        let accountPath = GLib.Variant.new('s', account.get_object_path());
-
-        let savedChannels = savedChannels.filter(function(a) {
-            return !a.account.equal(accountPath);
-        });
-        this._settings.set_value('saved-channel-list',
-                                 GLib.Variant.new('aa{sv}', savedChannels));
-    },
-
-    _updateAccountName: function(account, name, callback) {
-        let sv = { account: GLib.Variant.new('s', name) };
-        let asv = GLib.Variant.new('a{sv}', sv);
-        account.update_parameters_vardict_async(asv, [], callback);
-    },
-
-    _updateAccountServer: function(account, server, callback) {
-        let sv = { server: GLib.Variant.new('s', server.address),
-                   port: GLib.Variant.new('u', server.port) };
-        let asv = GLib.Variant.new('a{sv}', sv);
-        account.update_parameters_vardict_async(asv, [], callback);
-    },
-
-    _requestChannel: function(accountPath, targetType, targetId, time, callback) {
-        // have this in AccountMonitor?
-        let factory = Tp.AccountManager.dup().get_factory();
-        let account = factory.ensure_account(accountPath, []);
-
-        if (!account.enabled) {
-            // if we are requesting a channel for a disabled account, we
-            // are restoring saved channels; if the account has also never
-            // been online, it was removed since the channel was saved
-            if (!account.has_been_online)
-                this._removeSavedChannelsForAccount(account);
-            return;
-        }
-
-        if (!this._networkMonitor.network_available)
-            return;
-
-        let roomId = Polari.create_room_id(account,  targetId, targetType);
-
-        let params = account.dup_parameters_vardict().deep_unpack();
-        let server = params['server'].deep_unpack();
-        let accountServers = [];
-
-        // If predefined network, get alternate servers list
-        if (this._networksManager.getAccountIsPredefined(account))
-            accountServers = this._networksManager.getNetworkServers(account.service);
-
-        let requestData = {
-          account: account,
-          targetHandleType: targetType,
-          targetId: targetId,
-          roomId: roomId,
-          cancellable: new Gio.Cancellable(),
-          time: time,
-          retry: 0,
-          originalNick: account.nickname,
-          callback: callback,
-          alternateServers: accountServers.filter(s => s != server)
-        };
-
-        this._pendingRequests[roomId] = requestData;
-
-        this._ensureChannel(requestData);
-    },
-
-    _ensureChannel: function(requestData) {
-        let account = requestData.account;
-
-        let req = Tp.AccountChannelRequest.new_text(account, requestData.time);
-        req.set_target_id(requestData.targetHandleType, requestData.targetId);
-        req.set_delegate_to_preferred_handler(true);
-        let preferredHandler = Tp.CLIENT_BUS_NAME_BASE + 'Polari';
-        req.ensure_and_observe_channel_async(preferredHandler, requestData.cancellable,
-                                 Lang.bind(this,
-                                           this._onEnsureChannel, requestData));
-    },
-
-    _retryNickRequest: function(requestData) {
-        let account = requestData.account;
-
-        // Try again with a different nick
-        let params = account.dup_parameters_vardict().deep_unpack();
-        let oldNick = params['account'].deep_unpack();
-        let nick = oldNick + '_';
-        this._updateAccountName(account, nick, Lang.bind(this,
-            function() {
-                this._ensureChannel(requestData);
-            }));
-    },
-
-    _retryServerRequest: function(requestData) {
-        let account = requestData.account;
-
-        // Try again with a alternate server
-        let server = requestData.alternateServers.shift();
-        this._updateAccountServer(account, server, Lang.bind(this,
-            function() {
-                this._ensureChannel(requestData);
-            }));
-    },
-
-    _onEnsureChannel: function(req, res, requestData) {
-        let account = req.account;
-        let channel = null;
-
-        try {
-            channel = req.ensure_and_observe_channel_finish(res);
-        } catch (e if e.matches(Tp.Error, Tp.Error.DISCONNECTED)) {
-            // If we receive a disconnect error and the network is unavailable,
-            // then the error is not specific to polari and polari will
-            // just be in offline state.
-            if (!this._networkMonitor.network_available)
-                return;
-
-            let error = account.connection_error;
-            switch (error) {
-                case ConnectionError.CANCELLED:
-                    break; // disconnected due to user request, ignore
-                case ConnectionError.ALREADY_CONNECTED:
-                    if (requestData.retry++ < MAX_RETRIES) {
-                        this._retryNickRequest(requestData);
-                        return;
-                    }
-                    break;
-                case ConnectionError.NETWORK_ERROR:
-                case ConnectionError.NOT_AVAILABLE:
-                case ConnectionError.DISCONNECTED:
-                case ConnectionError.SERVICE_BUSY:
-                    if (requestData.alternateServers.length > 0) {
-                        this._retryServerRequest(requestData);
-                        return;
-                    }
-                default:
-                    Utils.debug('Account %s disconnected with error %s'.format(
-                                account.get_path_suffix(),
-                                error.replace(Tp.ERROR_PREFIX + '.', '')));
-            }
-        } catch (e if e.matches(Tp.Error, Tp.Error.CANCELLED)) {
-            // interrupted by user request, don't log
-        } catch (e) {
-            Utils.debug('Failed to ensure channel: ' + e.message);
-        }
-
-        if (requestData.callback)
-            requestData.callback(channel);
-
-        if (requestData.retry > 0)
-            this._updateAccountName(account, requestData.originalNick, null);
-        delete this._pendingRequests[requestData.roomId];
+        if (!this._window || present)
+            this.activate();
     },
 
     _onJoinRoom: function(action, parameter) {
         let [accountPath, channelName, time] = parameter.deep_unpack();
-        this._requestChannel(accountPath, Tp.HandleType.ROOM,
-                             channelName, time);
-
-        let factory = Tp.AccountManager.dup().get_factory();
-        let account = factory.ensure_account(accountPath, []);
-        this._addSavedChannel(account, channelName);
+        this._maybePresent(time);
     },
 
     _onMessageUser: function(action, parameter) {
         let [accountPath, contactName, message, time] = parameter.deep_unpack();
-        this._requestChannel(accountPath, Tp.HandleType.CONTACT,
-                             contactName, time, Lang.bind(this, this._sendMessage, message));
+        this._maybePresent(time);
     },
 
-    _sendMessage: function(channel, message) {
-        if (!message || !channel)
+    _ensureRetryData: function(account) {
+        let data = this._retryData.get(account.object_path);
+        if (data)
+            return data;
+
+        let params = account.dup_parameters_vardict().deep_unpack();
+        let server = params['server'].deep_unpack();
+        let accountName = params['account'].deep_unpack();
+        Utils.debug('Failed to connect to %s with username %s'.format(server, accountName));
+
+        let accountServers = [];
+        if (this._networksManager.getAccountIsPredefined(account))
+            accountServers = this._networksManager.getNetworkServers(account.service);
+
+        data = {
+            retry: 0,
+            originalAccountName: accountName,
+            alternateServers: accountServers.filter(s => s.address != server)
+        };
+        this._retryData.set(account.object_path, data);
+        return data;
+    },
+
+    _restoreAccountName: function(account) {
+        let data = this._retryData.get(account.object_path);
+        if (!data || !data.retry || !data.originalAccountName)
             return;
 
-        let TpMessage = Tp.ClientMessage.new_text(Tp.ChannelTextMessageType.NORMAL,
-                                                  message);
-        channel.send_message_async(TpMessage, 0, Lang.bind(this,
-            function(c, res) {
-            try {
-                c.send_message_finish(res);
-            } catch(e) {
-                // TODO: propagate to user
-                logError(e, 'Failed to send message')
+        let params = { account: new GLib.Variant('s', data.originalAccountName) };
+        let asv = new GLib.Variant('a{sv}', params);
+        account.update_parameters_vardict_async(asv, [], null);
+        delete data.originalAccountName;
+    },
+
+    _retryWithParams: function(account, params) {
+        account.update_parameters_vardict_async(params, [], () => {
+            let presence = Tp.ConnectionPresenceType.AVAILABLE;
+            let msg = account.requested_status_message;
+            account.request_presence_async(presence, 'available', msg, null);
+        });
+    },
+
+    _retryNickRequest: function(account) {
+        let retryData = this._ensureRetryData(account);
+
+        if (retryData.retry++ >= MAX_RETRIES)
+            return false;
+
+        let oldParams = account.dup_parameters_vardict().deep_unpack();
+        let nick = oldParams['account'].deep_unpack();
+
+        Utils.debug('Retrying with nickname %s'.format(nick + '_'));
+        let params = { account: new GLib.Variant('s', nick + '_') };
+        this._retryWithParams(account, new GLib.Variant('a{sv}', params));
+        return true;
+    },
+
+    _retryServerRequest: function(account) {
+        let retryData = this._ensureRetryData(account);
+
+        let server = retryData.alternateServers.shift();
+        if (!server)
+            return false;
+
+        Utils.debug('Retrying with %s:%d'.format(server.address, server.port));
+        let params = { server: new GLib.Variant('s', server.address),
+                       port: new GLib.Variant('u', server.port),
+                       'use-ssl': new GLib.Variant('b', server.ssl) };
+        this._retryWithParams(account, new GLib.Variant('a{sv}', params));
+        return true;
+    },
+
+    _onAccountStatusChanged: function(mon, account) {
+        let status = account.connection_status;
+
+        if (status == Tp.ConnectionStatus.CONNECTING)
+            return;
+
+        if (status == Tp.ConnectionStatus.DISCONNECTED) {
+            let reason = account.connection_status_reason;
+
+            if (reason == Tp.ConnectionStatusReason.NAME_IN_USE)
+                if (this._retryNickRequest(account))
+                    return;
+
+            if (reason == Tp.ConnectionStatusReason.NETWORK_ERROR ||
+                reason == Tp.ConnectionStatusReason.NONE_SPECIFIED)
+                if (this._retryServerRequest(account))
+                    return;
+
+            if (reason != Tp.ConnectionStatusReason.REQUESTED) {
+                let strReasons = Object.keys(Tp.ConnectionStatusReason);
+                Utils.debug('Account %s disconnected with reason %s'.format(
+                            account.display_name, strReasons[reason]));
             }
-        }));
-    },
-
-    _onLeaveRoom: function(action, parameter) {
-        let [roomId, message] = parameter.deep_unpack();
-        let reason = Tp.ChannelGroupChangeReason.NONE;
-        let room = this._chatroomManager.getRoomById(roomId);
-        if (!room)
-            return;
-        if (this._pendingRequests[roomId]) {
-            this._pendingRequests[roomId].cancellable.cancel();
-        } else if (room.channel) {
-            if (!message.length)
-                message = _("Good Bye"); // TODO - our first setting?
-            room.channel.leave_async(reason, message, Lang.bind(this,
-                function(c, res) {
-                    try {
-                        c.leave_finish(res);
-                    } catch(e) {
-                        logError(e, 'Failed to leave channel');
-                    }
-                }));
         }
-        this._removeSavedChannel(room.account, room.channel_name);
+
+        this._restoreAccountName(account);
     },
 
     _onLeaveCurrentRoom: function() {
-        let room = this._chatroomManager.getActiveRoom();
+        let room = this._window.active_room;
         if (!room)
             return;
         let action = this.lookup_action('leave-room');
         action.activate(GLib.Variant.new('(ss)', [room.id, '']));
     },
 
+    _onConnectAccount: function(action, parameter) {
+        let accountPath = parameter.deep_unpack();
+        let account = this._accountsMonitor.lookupAccount(accountPath);
+        if (account)
+            this._restoreAccountName(account);
+        this._retryData.delete(accountPath);
+    },
+
     _onToggleAction: function(action) {
         let state = action.get_state();
         action.change_state(GLib.Variant.new('b', !state.get_boolean()));
@@ -561,8 +437,7 @@ const Application = new Lang.Class({
 
     _onRemoveConnection: function(action, parameter){
         let accountPath = parameter.deep_unpack();
-        let factory = Tp.AccountManager.dup().get_factory();
-        let account = factory.ensure_account(accountPath, []);
+        let account = this._accountsMonitor.lookupAccount(accountPath);
         account.set_enabled_async(false, Lang.bind(this,
             function() {
                 let label = _("%s removed.").format(account.display_name);
@@ -584,8 +459,7 @@ const Application = new Lang.Class({
 
     _onEditConnection: function(action, parameter) {
         let accountPath = parameter.deep_unpack();
-        let factory = Tp.AccountManager.dup().get_factory();
-        let account = factory.ensure_account(accountPath, []);
+        let account = this._accountsMonitor.lookupAccount(accountPath);
         let dialog = new Connections.ConnectionProperties(account);
         dialog.transient_for = this._window;
         dialog.connect('response', Lang.bind(this,
