@@ -45,6 +45,9 @@ struct _PolariRoomPrivate {
   guint self_contact_notify_id;
   guint invalidated_id;
   guint group_contacts_changed_id;
+  guint message_sent_id;
+
+  gboolean ignore_identify;
 
   TpProxySignalConnection *properties_changed_id;
 };
@@ -78,6 +81,8 @@ enum
 
   MEMBERS_CHANGED,
 
+  IDENTIFY_SENT,
+
   LAST_SIGNAL
 };
 
@@ -85,6 +90,8 @@ static guint signals[LAST_SIGNAL];
 
 static GRegex *color_code_regex = NULL;
 
+static GRegex *identify_message_regex = NULL;
+
 G_DEFINE_TYPE_WITH_PRIVATE (PolariRoom, polari_room, G_TYPE_OBJECT)
 
 static void polari_room_set_channel (PolariRoom *room, TpChannel *channel);
@@ -237,6 +244,80 @@ polari_room_remove_member (PolariRoom *room,
   }
 }
 
+static void
+on_identify_message_sent (GObject      *source,
+                          GAsyncResult *result,
+                          gpointer      user_data)
+{
+  TpTextChannel *channel = TP_TEXT_CHANNEL (source);
+  GTask *task = user_data;
+  PolariRoom *room = g_task_get_source_object (task);
+  GError *error = NULL;
+
+  if (!tp_text_channel_send_message_finish (channel, result, NULL, &error))
+    {
+      room->priv->ignore_identify = FALSE;
+
+      g_task_return_error (task, error);
+      g_object_unref (task);
+      return;
+    }
+
+  g_task_return_boolean (task, TRUE);
+  g_object_unref (task);
+}
+
+void
+polari_room_send_identify_message_async (PolariRoom          *room,
+                                         const char          *username,
+                                         const char          *password,
+                                         GAsyncReadyCallback  callback,
+                                         gpointer             user_data)
+{
+  PolariRoomPrivate *priv;
+  TpMessage *message;
+  GTask *task;
+  char *text;
+
+  g_return_if_fail (POLARI_IS_ROOM (room));
+  g_return_if_fail (username != NULL && password != NULL);
+
+  priv = room->priv;
+
+  task = g_task_new (room, NULL, callback, user_data);
+
+  if (priv->channel == NULL)
+    {
+      g_task_return_new_error (task, G_IO_ERROR, G_IO_ERROR_NOT_CONNECTED,
+                               "The room is disconnected.");
+      g_object_unref (task);
+      return;
+    }
+
+  /* Don't emit ::identify-sent for our own identify message */
+  room->priv->ignore_identify = TRUE;
+
+  text = g_strdup_printf ("identify %s %s", username, password);
+  message = tp_client_message_new_text (TP_CHANNEL_TEXT_MESSAGE_TYPE_NORMAL, text);
+
+  tp_text_channel_send_message_async (TP_TEXT_CHANNEL (priv->channel), message,
+                                      0, on_identify_message_sent, task);
+
+  g_object_unref (message);
+  g_free (text);
+}
+
+gboolean
+polari_room_send_identify_message_finish (PolariRoom    *room,
+                                          GAsyncResult  *result,
+                                          GError       **error)
+{
+  g_return_val_if_fail (POLARI_IS_ROOM (room), FALSE);
+  g_return_val_if_fail (g_task_is_valid (result, room), FALSE);
+
+  return g_task_propagate_boolean (G_TASK (result), error);
+}
+
 static char *
 strip_color_codes (const char *string) {
   if (G_UNLIKELY (color_code_regex == NULL))
@@ -378,6 +459,45 @@ on_group_contacts_changed (TpChannel  *channel,
   g_free (message);
 }
 
+static void
+on_message_sent (TpTextChannel      *channel,
+                 TpSignalledMessage *message,
+                 guint               flags,
+                 char               *token,
+                 gpointer            user_data)
+{
+  PolariRoom *room = user_data;
+  PolariRoomPrivate *priv = room->priv;
+  GMatchInfo *match;
+  char *text, *stripped_text;
+
+  if (priv->type != TP_HANDLE_TYPE_CONTACT)
+    return;
+
+  text = tp_message_to_text (TP_MESSAGE (message), NULL);
+  stripped_text = g_strstrip (text);
+
+  if (G_UNLIKELY (identify_message_regex == NULL))
+    identify_message_regex = g_regex_new ("^identify (?:(\\w+) )?(\\S+)$",
+                                                G_REGEX_OPTIMIZE, 0, NULL);
+  if (g_regex_match (identify_message_regex, stripped_text, 0, &match))
+    {
+      char *username = g_match_info_fetch (match, 1);
+      char *password = g_match_info_fetch (match, 2);
+
+      if (!priv->ignore_identify)
+        g_signal_emit (room, signals[IDENTIFY_SENT], 0, username, password);
+
+      priv->ignore_identify = FALSE;
+
+      g_free (username);
+      g_free (password);
+    }
+
+  g_match_info_free (match);
+  g_free (text);
+}
+
 static void
 on_channel_invalidated (TpProxy  *channel,
                         guint     domain,
@@ -575,6 +695,7 @@ polari_room_set_channel (PolariRoom *room,
   if (priv->channel)
     {
       g_signal_handler_disconnect (priv->channel, priv->group_contacts_changed_id);
+      g_signal_handler_disconnect (priv->channel, priv->message_sent_id);
       g_signal_handler_disconnect( priv->channel, priv->invalidated_id);
       g_signal_handler_disconnect (tp_channel_get_connection (priv->channel),
                                    priv->self_contact_notify_id);
@@ -610,6 +731,9 @@ polari_room_set_channel (PolariRoom *room,
       priv->group_contacts_changed_id =
         g_signal_connect (channel, "group-contacts-changed",
                           G_CALLBACK (on_group_contacts_changed), room);
+      priv->message_sent_id =
+        g_signal_connect ( channel, "message-sent",
+                          G_CALLBACK (on_message_sent), room);
       priv->invalidated_id =
         g_signal_connect (channel, "invalidated",
                           G_CALLBACK (on_channel_invalidated), room);
@@ -848,6 +972,14 @@ polari_room_class_init (PolariRoomClass *klass)
                   0,
                   NULL, NULL, NULL,
                   G_TYPE_NONE, 0);
+
+  signals[IDENTIFY_SENT] =
+    g_signal_new ("identify-sent",
+                  G_TYPE_FROM_CLASS (klass),
+                  G_SIGNAL_RUN_LAST,
+                  0,
+                  NULL, NULL, NULL,
+                  G_TYPE_NONE, 2, G_TYPE_STRING, G_TYPE_STRING);
 }
 
 static void
