@@ -7,8 +7,8 @@ const Pango = imports.gi.Pango;
 const Tp = imports.gi.TelepathyGLib;
 
 const AccountsMonitor = imports.accountsMonitor;
-const ChatroomManager = imports.chatroomManager;
 const Lang = imports.lang;
+const RoomManager = imports.roomManager;
 
 function _onPopoverVisibleChanged(popover) {
     let context = popover.relative_to.get_style_context();
@@ -151,6 +151,7 @@ const RoomListHeader = new Lang.Class({
                        'popoverStatus',
                        'popoverTitle',
                        'popoverPassword',
+                       'popoverConnect',
                        'popoverReconnect',
                        'popoverRemove',
                        'popoverProperties',
@@ -173,6 +174,7 @@ const RoomListHeader = new Lang.Class({
             }));
 
         let target = new GLib.Variant('o', this._account.get_object_path());
+        this._popoverConnect.action_target = target;
         this._popoverReconnect.action_target = target;
         this._popoverRemove.action_target = target;
         this._popoverProperties.action_target = target;
@@ -197,17 +199,23 @@ const RoomListHeader = new Lang.Class({
                                   Lang.bind(this, this._onConnectionStatusChanged));
         this._onConnectionStatusChanged();
 
+        let presenceChangedId =
+            this._account.connect('notify::requested-presence-type',
+                                  Lang.bind(this, this._onRequestedPresenceChanged));
+        this._onRequestedPresenceChanged();
+
         this.connect('destroy', Lang.bind(this, function() {
             this._account.disconnect(displayNameChangedId);
             this._account.disconnect(connectionStatusChangedId);
+            this._account.disconnect(presenceChangedId);
         }));
     },
 
     _onDisplayNameChanged: function() {
         this._label.label = this._account.display_name;
 
-	/* update pop-over status label */
-	this._onConnectionStatusChanged();
+        /* update pop-over status label */
+        this._onConnectionStatusChanged();
 
         let parent;
         do
@@ -279,6 +287,13 @@ const RoomListHeader = new Lang.Class({
         }
     },
 
+    _onRequestedPresenceChanged: function() {
+        let presence = this._account.requested_presence_type;
+        let offline = presence == Tp.ConnectionPresenceType.OFFLINE;
+        this._popoverConnect.visible = offline;
+        this._popoverReconnect.visible = !offline;
+    },
+
     _getStatusLabel: function() {
         switch (this._account.connection_status) {
             case Tp.ConnectionStatus.CONNECTED:
@@ -335,105 +350,75 @@ const RoomList = new Lang.Class({
         this.set_header_func(Lang.bind(this, this._updateHeader));
         this.set_sort_func(Lang.bind(this, this._sort));
 
-        this._placeholders = {};
-        this._roomRows = {};
+        this._placeholders = new Map();
+        this._roomRows = new Map();
 
         this._accountsMonitor = AccountsMonitor.getDefault();
-        this._accountsMonitor.connect('account-manager-prepared', Lang.bind(this,
-            function(mon, am) {
-                let accounts = this._accountsMonitor.dupAccounts();
-                for (let i = 0; i < accounts.length; i++)
-                    this._accountAdded(mon, accounts[i]);
-
-                am.connect('account-enabled', Lang.bind(this,
-                    function(am, account) {
-                        this._updatePlaceholderVisibility(account);
-                    }));
-                am.connect('account-disabled', Lang.bind(this,
-                    function(am, account) {
-                        this._updatePlaceholderVisibility(account);
-                    }));
-            }));
         this._accountsMonitor.connect('account-added',
                                       Lang.bind(this, this._accountAdded));
         this._accountsMonitor.connect('account-removed',
                                       Lang.bind(this, this._accountRemoved));
+        this._accountsMonitor.connect('account-enabled', (mon, account) => {
+            this._updatePlaceholderVisibility(account);
+        });
+        this._accountsMonitor.connect('account-disabled', (mon, account) => {
+            this._updatePlaceholderVisibility(account);
+        });
+        this._accountsMonitor.prepare(() => {
+            this._accountsMonitor.accounts.forEach(account => {
+                this._accountAdded(this._accountsMonitor, account);
+            });
+        });
 
-        this._roomManager = ChatroomManager.getDefault();
+        this._roomManager = RoomManager.getDefault();
         this._roomManager.connect('room-added',
                                   Lang.bind(this, this._roomAdded));
         this._roomManager.connect('room-removed',
                                   Lang.bind(this, this._roomRemoved));
-        this._roomManager.connect('active-changed',
-                                  Lang.bind(this, this._activeRoomChanged));
+        this._roomManager.rooms.forEach(r => { this._roomAdded(this._roomManager, r); });
 
         let app = Gio.Application.get_default();
-        this._leaveAction = app.lookup_action('leave-room');
-        this._leaveAction.connect('activate',
-                                  Lang.bind(this, this._onLeaveActivated));
-
-        let action;
-        action = app.lookup_action('next-room');
-        action.connect('activate', Lang.bind(this,
-            function() {
-                this._moveSelection(Gtk.DirectionType.DOWN);
-            }));
-        action = app.lookup_action('previous-room');
-        action.connect('activate', Lang.bind(this,
-            function() {
-                this._moveSelection(Gtk.DirectionType.UP);
-            }));
-        action = app.lookup_action('first-room');
-        action.connect('activate', Lang.bind(this,
-            function() {
-                let row = this._getRoomRowAtIndex(0);
-                if (row)
-                    this.select_row(row);
-            }));
-        action = app.lookup_action('last-room');
-        action.connect('activate', Lang.bind(this,
-            function() {
-                let nRows = this._roomManager.roomCount;
-                let row = this._getRoomRowAtIndex(nRows - 1);
-                if (row)
-                    this.select_row(row);
-            }));
-        action = app.lookup_action('nth-room');
-        action.connect('activate', Lang.bind(this,
-            function(action, param) {
-                let n = param.get_int32();
-                if (n > this._roomManager.roomCount)
-                    return;
-                this.select_row(this._getRoomRowAtIndex(n - 1));
-            }));
-        action = app.lookup_action('next-pending-room');
-        action.connect('activate', Lang.bind(this,
-            function() {
-                this._moveSelectionFull(Gtk.DirectionType.DOWN,
-                                        (row) => { return row.hasPending; });
-            }));
-        action = app.lookup_action('previous-pending-room');
-        action.connect('activate', Lang.bind(this,
-            function() {
-                this._moveSelectionFull(Gtk.DirectionType.UP,
-                                        (row) => { return row.hasPending; });
-            }));
+        let actions = [
+            { name: 'next-room',
+              handler: () => { this._moveSelection(Gtk.DirectionType.DOWN); } },
+            { name: 'previous-room',
+              handler: () => { this._moveSelection(Gtk.DirectionType.UP); } },
+            { name: 'first-room',
+              handler: () => { this._selectRoomAtIndex(0); } },
+            { name: 'last-room',
+              handler: () => {
+                  let nRows = this._roomManager.roomCount;
+                  this._selectRoomAtIndex(nRows - 1);
+              } },
+            { name: 'nth-room',
+              handler: (a, param) => {
+                  this._selectRoomAtIndex(param.get_int32() - 1);
+              } },
+            { name: 'next-pending-room',
+              handler: () => { this._moveSelectionFull(Gtk.DirectionType.DOWN,
+                                                       row => row.hasPending); } },
+            { name: 'previous-pending-room',
+              handler: () => { this._moveSelectionFull(Gtk.DirectionType.UP,
+                                                       row => row.hasPending); } }
+        ];
+        actions.forEach(a => {
+            app.lookup_action(a.name).connect('activate', a.handler);
+        });
     },
 
-    _onLeaveActivated: function(action, param) {
-        let [id, ] = param.deep_unpack();
-        let row = this._roomRows[id];
+    vfunc_realize: function() {
+        this.parent();
 
-        this._moveSelectionFromRow(row);
-        row.hide();
+        let toplevel = this.get_toplevel();
+        toplevel.connect('notify::active-room',
+                         Lang.bind(this, this._activeRoomChanged));
+        this._activeRoomChanged();
     },
 
     _rowToRoomIndex: function(index) {
-        let nPlaceholders = Object.keys(this._placeholders).filter(
-            Lang.bind(this, function(a) {
-                return this._placeholders[a].get_index() < index;
-            })).length;
-        return index - nPlaceholders;
+        let placeholders = [...this._placeholders.values()];
+        let nBefore = placeholders.filter(p => p.get_index() < index).length;
+        return index - nBefore;
     },
 
     _roomToRowIndex: function(index) {
@@ -448,6 +433,12 @@ const RoomList = new Lang.Class({
         return this.get_row_at_index(this._roomToRowIndex(index));
     },
 
+    _selectRoomAtIndex: function(index) {
+        let row = this._getRoomRowAtIndex(index);
+        if (row)
+            this.select_row(row);
+    },
+
     _moveSelection: function(direction) {
         this._moveSelectionFull(direction, () => { return true; });
     },
@@ -475,8 +466,8 @@ const RoomList = new Lang.Class({
         if (this._roomManager.roomCount == 0)
             return;
 
-        let activeRoom = this._roomManager.getActiveRoom();
-        let current = this._roomRows[activeRoom.id];
+        let toplevel = this.get_toplevel();
+        let current = this._roomRows.get(toplevel.active_room.id);
 
         if (current != row)
             return;
@@ -492,21 +483,22 @@ const RoomList = new Lang.Class({
         let newSelected = this.get_selected_row();
         if (newSelected != row)
             newActive = newSelected.room;
-        this._roomManager.setActiveRoom(newActive);
+        toplevel.active_room = newActive;
 
         if (selected != row)
             this.select_row(selected);
     },
 
     _accountAdded: function(am, account) {
-        if (this._placeholders[account])
+        if (this._placeholders.has(account))
             return;
 
         let placeholder = new Gtk.ListBoxRow({ selectable: false,
-                                               activatable: false });
+                                               activatable: false,
+                                               no_show_all: true });
         placeholder.account = account;
 
-        this._placeholders[account] = placeholder;
+        this._placeholders.set(account, placeholder);
         this.add(placeholder);
 
         placeholder.connect('notify::visible', Lang.bind(this,
@@ -518,56 +510,54 @@ const RoomList = new Lang.Class({
     },
 
     _accountRemoved: function(am, account) {
-        let placeholder = this._placeholders[account];
+        let placeholder = this._placeholders.get(account);
 
         if (!placeholder)
             return;
 
-        delete this._placeholders[account];
+        this._placeholders.delete(account);
         placeholder.destroy();
     },
 
     _roomAdded: function(roomManager, room) {
+        if (this._roomRows.has(room.id))
+            return;
+
         let row = new RoomRow(room);
         this.add(row);
-        this._roomRows[room.id] = row;
+        this._roomRows.set(room.id, row);
 
-        row.connect('destroy', Lang.bind(this,
-            function(w) {
-                delete this._roomRows[w.room.id];
-            }));
-        this._placeholders[room.account].hide();
+        row.connect('destroy', w => { this._roomRows.delete(w.room.id); });
+        this._placeholders.get(room.account).hide();
     },
 
     _roomRemoved: function(roomManager, room) {
-        let row = this._roomRows[room.id];
+        let row = this._roomRows.get(room.id);
         if (!row)
             return;
 
         this._moveSelectionFromRow(row);
         row.destroy();
-        delete this._roomRows[room.id];
+        this._roomRows.delete(room.id);
         this._updatePlaceholderVisibility(room.account);
     },
 
     _updatePlaceholderVisibility: function(account) {
         if (!account.enabled) {
-            this._placeholders[account].hide();
+            this._placeholders.get(account).hide();
             return;
         }
 
-        let ids = Object.keys(this._roomRows);
-        let hasRooms = ids.some(Lang.bind(this,
-            function(id) {
-                return this._roomRows[id].account == account;
-            }));
-        this._placeholders[account].visible = !hasRooms;
+        let rows = [...this._roomRows.values()];
+        let hasRooms = rows.some(r => r.account == account);
+        this._placeholders.get(account).visible = !hasRooms;
     },
 
-    _activeRoomChanged: function(roomManager, room) {
+    _activeRoomChanged: function() {
+        let room = this.get_toplevel().active_room;
         if (!room)
             return;
-        let row = this._roomRows[room.id];
+        let row = this._roomRows.get(room.id);
         if (!row)
             return;
 
@@ -577,7 +567,7 @@ const RoomList = new Lang.Class({
     },
 
     vfunc_row_selected: function(row) {
-        this._roomManager.setActiveRoom(row ? row.room : null);
+        this.get_toplevel().active_room = row ? row.room : null;
         if (row)
             row.selected();
     },
@@ -608,8 +598,8 @@ const RoomList = new Lang.Class({
         let account1 = row1.account;
         let account2 = row2.account;
 
-        let hasRooms1 = !this._placeholders[account1].visible;
-        let hasRooms2 = !this._placeholders[account2].visible;
+        let hasRooms1 = !this._placeholders.get(account1).visible;
+        let hasRooms2 = !this._placeholders.get(account2).visible;
 
         if (hasRooms1 != hasRooms2)
             return hasRooms1 ? -1 : 1;
